@@ -4,18 +4,17 @@ import asyncio
 import os
 import shutil
 import time
-import warnings
 from pathlib import Path
 from urllib.parse import urlparse
 
 import aiofiles
 import httpx
-import tqdm
 from httpx._types import HeaderTypes
 
 from throttlebuster.constants import (
     CURRENT_WORKING_DIR,
     DEFAULT_CHUNK_SIZE,
+    DEFAULT_READ_TIMEOUT_ATTEMPTS,
     DEFAULT_REQUEST_HEADERS,
     DEFAULT_TASKS,
     DEFAULT_TASKS_LIMIT,
@@ -28,19 +27,15 @@ from throttlebuster.exceptions import (
     IncompatibleServerError,
 )
 from throttlebuster.helpers import (
+    CustomTqdm,
     DownloadUtils,
     assert_instance,
-    get_duration_string,
     get_filesize_string,
     logger,
-    loop,
     sanitize_filename,
 )
 from throttlebuster.models import DownloadedFile, DownloadTracker
 
-warnings.simplefilter("ignore", category=(tqdm.std.TqdmWarning,))
-"""Raised due to frac*"""
-
 
 class ThrottleBuster(DownloadUtils):
     """Performs file download using multiple tasks in attempt
@@ -185,7 +180,7 @@ class ThrottleBuster(DownloadUtils):
     async def _downloader(
         self,
         download_tracker: DownloadTracker,
-        progress_bar: tqdm.tqdm,
+        progress_bar: CustomTqdm,
         progress_hook: callable,
     ) -> DownloadTracker:
         """Downloads each file part"""
@@ -206,7 +201,9 @@ class ThrottleBuster(DownloadUtils):
             downloaded_size = os.path.getsize(download_tracker.saved_to)
             download_tracker.bytes_offset += downloaded_size
             download_tracker.update_downloaded_size(downloaded_size)
-            progress_bar.total -= self.bytes_to_mb(downloaded_size)
+            progress_bar.n += self.bytes_to_mb(downloaded_size)
+            progress_bar.last_print_t = time.time()
+            progress_bar.last_print_n = progress_bar.n
 
         else:
             download_tracker.download_mode = DownloadMode.START
@@ -250,6 +247,8 @@ class ThrottleBuster(DownloadUtils):
         file_size: int = None,
         keep_parts: bool = False,
         suppress_incompatible_error: bool = False,
+        timeout_retry_attempts: int = DEFAULT_READ_TIMEOUT_ATTEMPTS,
+        retry_attempts_count: int = 0,
         colour: str = "cyan",
         simple: bool = False,
         test: bool = False,
@@ -268,6 +267,7 @@ class ThrottleBuster(DownloadUtils):
             file_size (int, optional): Size of the file to be downloaded. Defaults to None.
             keep_parts (bool, optional): Whether to retain the separate download parts. Defaults to False.
             suppress_incompatible_error (bool, optional): Do no raise error when response headers lack Etag. Defaults to False.
+            timeout_retry_attempts (int, optional): Number of times to retry download upon read request timing out. Defaults to DEFAULT_READ_TIMEOUT_ATTEMPTS.
             leave (bool, optional): Keep all leaves of the progressbar. Defaults to True.
             colour (str, optional): Progress bar display color. Defaults to "cyan".
             simple (bool, optional): Show percentage and bar only in progressbar. Deafults to False.
@@ -295,134 +295,191 @@ class ThrottleBuster(DownloadUtils):
 
         logger.debug(f'Initializing download (tasks - {self.tasks}) for file in url - "{url}"')
 
-        async with self.client.stream("GET", url=url) as stream:
-            stream.raise_for_status()
+        try:
+            async with self.client.stream("GET", url=url) as stream:
+                stream.raise_for_status()
 
-            if (
-                stream.headers.get("Etag") is None
-                and suppress_incompatible_error is False
-                and self.tasks != 1
-            ):
-                raise IncompatibleServerError(
-                    "Server response header lacks Etag value which means "
-                    "it doesn't support resuming downloads. "
-                    "Set tasks to 1 or activate suppress_incompatible_error parameter "
-                    "to silence this error."
-                )
-
-            content_length = stream.headers.get("content-length", file_size)
-            if type(content_length) is str:
-                content_length = int(content_length)
-
-            filename = filename or self.get_filename_from_header(stream.headers)
-
-            if filename is None:
-                # Try to get from path
-                _, filename = os.path.split(urlparse(url).path)
-                if not filename:
-                    raise FilenameNotFoundError(
-                        "Unable to get filename. Pass value using parameter filename to suppress this error"
+                if (
+                    stream.headers.get("Etag") is None
+                    and suppress_incompatible_error is False
+                    and self.tasks != 1
+                ):
+                    raise IncompatibleServerError(
+                        "Server response header lacks Etag value which means "
+                        "it doesn't support resuming downloads. "
+                        "Set tasks to 1 or activate suppress_incompatible_error parameter "
+                        "to silence this error."
                     )
 
-            filename = sanitize_filename(filename)
+                content_length = stream.headers.get("content-length", file_size)
+                if type(content_length) is str:
+                    content_length = int(content_length)
 
-            final_saved_to = self._generate_saved_to(filename, self.dir)
+                filename = filename or self.get_filename_from_header(stream.headers)
 
-            if content_length is None:
-                raise FilesizeNotFoundError(
-                    "Unable to get the content-length of the file from server response. "
-                    "Set the content-length using parameter file_size to suppres this error."
-                )
+                if filename is None:
+                    # Try to get from path
+                    _, filename = os.path.split(urlparse(url).path)
+                    if not filename:
+                        raise FilenameNotFoundError(
+                            "Unable to get filename. Pass value using filename parameter "
+                            "to suppress this error"
+                        )
 
-            elif not test and final_saved_to.exists() and mode is not DownloadMode.START:
-                if os.path.getsize(final_saved_to) == content_length:
-                    logger.warning(f'Download already completed for the file in path "{final_saved_to}"')
-                    return DownloadedFile(
-                        url=url,
-                        saved_to=final_saved_to,
-                        size=os.path.getsize(final_saved_to),
-                        duration=0,
-                        file_parts=[],
-                        merge_duration=0,
-                        expected_size=content_length,
+                filename = sanitize_filename(filename)
+
+                final_saved_to = self._generate_saved_to(filename, self.dir)
+
+                if content_length is None:
+                    raise FilesizeNotFoundError(
+                        "Unable to get the content-length of the file from server response. "
+                        "Set the content-length using parameter file_size to suppres this error."
                     )
 
-            size_with_unit = get_filesize_string(content_length)
-            filename_disp = filename if len(filename) <= 8 + 3 else filename[:8] + "..."
+                elif not test and final_saved_to.exists() and mode is not DownloadMode.START:
+                    if os.path.getsize(final_saved_to) == content_length:
+                        logger.warning(f'Download already completed for the file in path "{final_saved_to}"')
+                        return DownloadedFile(
+                            url=url,
+                            saved_to=final_saved_to,
+                            size=os.path.getsize(final_saved_to),
+                            duration=0,
+                            file_parts=[],
+                            merge_duration=0,
+                            expected_size=content_length,
+                        )
 
-            if test:
-                logger.info(f"Download test passed successfully ({size_with_unit}) - {final_saved_to}")
-                return stream
+                size_with_unit = get_filesize_string(content_length)
+                filename_disp = filename if len(filename) <= 8 + 3 else filename[:8] + "..."
 
-            p_bar = tqdm.tqdm(
-                total=self.bytes_to_mb(content_length),
-                desc=f"Downloading{f' [{filename_disp}]'}",
-                unit="Mb",
-                disable=disable_progress_bar,
-                colour=colour,
-                leave=leave,
-                ascii=ascii,
-                bar_format=(
-                    "{l_bar}{bar} | %(size)s" % (dict(size=size_with_unit))
-                    if simple
-                    else "{l_bar}{bar}{r_bar}"
-                ),
-                **kwargs,
-            )
+                if test:
+                    logger.info(f"Download test passed successfully ({size_with_unit}) - {final_saved_to}")
+                    return stream
 
-            for index, offset_load in enumerate(self.get_offset_load(content_length, self.tasks)):
-                offset, load = offset_load
-                download_tracker = DownloadTracker(
-                    url=url,
-                    saved_to=self._generate_saved_to(
-                        f"{filename}-{offset, load}{self.part_extension}",
-                        self.part_dir,
-                        index,
+                logger.info(
+                    f"{'Starting' if retry_attempts_count == 0 else 'Resuming'} "
+                    f'download process ({self.tasks} tasks, {size_with_unit}) - "{filename}"'
+                )
+                p_bar = CustomTqdm(
+                    total=self.bytes_to_mb(content_length),
+                    desc=f"Downloading{f' [{filename_disp}]'}",
+                    unit="Mb",
+                    disable=disable_progress_bar,
+                    colour=colour,
+                    leave=leave,
+                    ascii=ascii,
+                    bar_format=(
+                        "{l_bar}{bar} | %(size)s" % (dict(size=size_with_unit))
+                        if simple
+                        else "{l_bar}{bar}{r_bar}"
                     ),
-                    index=index,
-                    bytes_offset=offset,
-                    expected_size=load,
-                    download_mode=mode,
+                    **kwargs,
                 )
 
-                download_tracker_items.append(download_tracker)
-                async_task = asyncio.create_task(
-                    self._downloader(
-                        download_tracker,
-                        progress_bar=p_bar,
-                        progress_hook=progress_hook,
+                for index, offset_load in enumerate(self.get_offset_load(content_length, self.tasks)):
+                    offset, load = offset_load
+                    download_tracker = DownloadTracker(
+                        url=url,
+                        saved_to=self._generate_saved_to(
+                            f"{filename}-{offset}{self.part_extension}",
+                            self.part_dir,
+                            index,
+                        ),
+                        index=index,
+                        bytes_offset=offset,
+                        expected_size=load,
+                        download_mode=mode,
                     )
+
+                    download_tracker_items.append(download_tracker)
+                    async_task = asyncio.create_task(
+                        self._downloader(
+                            download_tracker,
+                            progress_bar=p_bar,
+                            progress_hook=progress_hook,
+                        )
+                    )
+                    async_task_items.append(async_task)
+
+                download_start_time = time.time()
+
+                file_parts = await asyncio.gather(*async_task_items)
+
+                download_duration = time.time() - download_start_time
+
+                merge_start_time = time.time()
+                p_bar.close()
+
+                saved_to = await self._merge_parts(file_parts, filename=filename, keep_parts=keep_parts)
+
+                downloaded_file = DownloadedFile(
+                    url=url,
+                    saved_to=saved_to,
+                    expected_size=content_length,
+                    size=os.path.getsize(saved_to),
+                    file_parts=file_parts,
+                    merge_duration=time.time() - merge_start_time,
+                    duration=download_duration,
                 )
-                async_task_items.append(async_task)
 
-            download_start_time = time.time()
+                logger.info(
+                    f"Done downloading {downloaded_file.size_string} "
+                    f"in {downloaded_file.duration_string} "
+                    f"{'merged' if self.tasks > 1 else 'moved'} in "
+                    f"{downloaded_file.merge_duration_string}, "
+                    f'saved to "{downloaded_file.saved_to}"'
+                )
 
-            logger.info(f"Starting download process (tasks - {self.tasks}) - {filename}")
+                return downloaded_file
 
-            file_parts = await asyncio.gather(*async_task_items)
-            download_duration = time.time() - download_start_time
+        except httpx.ReadTimeout as e:
+            retry_attempts_count += 1
 
-            merge_start_time = time.time()
-            saved_to = await self._merge_parts(file_parts, filename=filename, keep_parts=keep_parts)
+            if retry_attempts_count <= timeout_retry_attempts:
+                # Retry
 
-            downloaded_file = DownloadedFile(
-                url=url,
-                saved_to=saved_to,
-                expected_size=content_length,
-                size=os.path.getsize(saved_to),
-                file_parts=file_parts,
-                merge_duration=time.time() - merge_start_time,
-                duration=download_duration,
-            )
+                p_bar.close()
 
-            logger.info(
-                f"Done downloading {get_filesize_string(downloaded_file.size)} "
-                f"in {get_duration_string(downloaded_file.duration)} "
-                f'saved to "{downloaded_file.saved_to}"'
-            )
+                logger.info(
+                    f"Retrying download after read request timed out - "
+                    f"attempt number ({retry_attempts_count}/{timeout_retry_attempts})"
+                )
 
-            return downloaded_file
+                return await self.run(
+                    url=url,
+                    filename=filename,
+                    progress_hook=progress_hook,
+                    mode=DownloadMode.AUTO,  # Changed
+                    disable_progress_bar=disable_progress_bar,
+                    file_size=file_size,
+                    keep_parts=keep_parts,
+                    suppress_incompatible_error=suppress_incompatible_error,
+                    timeout_retry_attempts=timeout_retry_attempts,
+                    retry_attempts_count=retry_attempts_count,
+                    colour=colour,
+                    simple=simple,
+                    test=test,
+                    leave=leave,
+                    ascii=ascii,
+                    **kwargs,
+                )
+
+            else:
+                if timeout_retry_attempts:
+                    logger.warning(
+                        f"Giving up on download after exhausting all {timeout_retry_attempts} "
+                        "read timeout retry attempts."
+                    )
+
+                else:
+                    logger.info(
+                        "Download read request has timed out. In order to automatically retry the "
+                        "process, declare value for retry attempts using parameter "
+                        "timeout_retry_attempts"
+                    )
+
+                raise e
 
     def run_sync(self, *args, **kwargs) -> DownloadedFile | httpx.Response:
         """Synchronously initiate download process of a file."""
-        return loop.run_until_complete(self.run(*args, **kwargs))
+        return asyncio.run(self.run(*args, **kwargs))
